import torch

import numpy as np
from torch.utils.data import Dataset
from torchvision.transforms import ToPILImage, RandomCrop, RandomHorizontalFlip, ToTensor, Normalize, Compose

from autokeras.constant import Constant


class OneHotEncoder:
    """A class that can format data

    This class provide ways to transform data's classification into vector

    Attributes:
          data: the input data
          n_classes: the number of classification
          labels: the number of label
          label_to_vec: mapping from label to vector
          int_to_label: mapping from int to label
    """

    def __init__(self):
        """Init OneHotEncoder"""
        self.data = None
        self.n_classes = 0
        self.labels = None
        self.label_to_vec = {}
        self.int_to_label = {}

    def fit(self, data):
        """Create mapping from label to vector, and vector to label"""
        data = np.array(data).flatten()
        self.labels = set(data)
        self.n_classes = len(self.labels)
        for index, label in enumerate(self.labels):
            vec = np.array([0] * self.n_classes)
            vec[index] = 1
            self.label_to_vec[label] = vec
            self.int_to_label[index] = label

    def transform(self, data):
        """Get vector for every element in the data array"""
        data = np.array(data)
        if len(data.shape) > 1:
            data = data.flatten()
        return np.array(list(map(lambda x: self.label_to_vec[x], data)))

    def inverse_transform(self, data):
        """Get label for every element in data"""
        return np.array(list(map(lambda x: self.int_to_label[x], np.argmax(np.array(data), axis=1))))


class DataTransformer:
    def __init__(self, data, augment=Constant.DATA_AUGMENTATION):
        self.max_val = data.max()
        data = data / self.max_val
        self.mean = np.mean(data, axis=(0, 1, 2), keepdims=True).flatten()
        self.std = np.std(data, axis=(0, 1, 2), keepdims=True).flatten()
        self.augment = augment

    def transform_train(self, data, targets=None):
        data = data / self.max_val
        data = torch.Tensor(data.transpose(0, 3, 1, 2))
        if not self.augment:
            augment_list = []
        else:
            augment_list = [ToPILImage(),
                            RandomCrop(data.shape[2:], padding=4),
                            RandomHorizontalFlip(),
                            ToTensor()
                            ]
        common_list = [Normalize(torch.Tensor(self.mean), torch.Tensor(self.std))]
        data_transforms = Compose(augment_list + common_list)
        return MultiTransformDataset(data, targets, data_transforms)

    def transform_test(self, data, targets=None):
        data = data / self.max_val
        data = torch.Tensor(data.transpose(0, 3, 1, 2))
        common_list = [Normalize(torch.Tensor(self.mean), torch.Tensor(self.std))]
        data_transforms = Compose(common_list)
        return MultiTransformDataset(data, targets, data_transforms)


class MultiTransformDataset(Dataset):
    def __init__(self, dataset, target, compose):
        self.dataset = dataset
        self.target = target
        self.compose = compose

    def __getitem__(self, index):
        feature = self.dataset[index]
        if self.target is None:
            return self.compose(feature)
        return self.compose(feature), self.target[index]

    def __len__(self):
        return len(self.dataset)



class Constant:
    DEFAULT_SAVE_PATH = '/tmp/autokeras/'

    # Data

    VALIDATION_SET_SIZE = 0.08333

    # Searcher

    MAX_MODEL_NUM = 1000
    BETA = 10.576
    KERNEL_LAMBDA = 1.0
    T_MIN = 0.0001
    N_NEIGHBOURS = 8
    # T_MIN = 0.8
    # N_NEIGHBOURS = 1

    # Model Defaults

    DENSE_DROPOUT_RATE = 0.5
    CONV_DROPOUT_RATE = 0.25
    CONV_BLOCK_DISTANCE = 3
    DENSE_BLOCK_DISTANCE = 2
    MODEL_LEN = 3
    MODEL_WIDTH = 64

    # ModelTrainer

    DATA_AUGMENTATION = True
    MAX_ITER_NUM = 200
    MIN_LOSS_DEC = 1e-4
    MAX_NO_IMPROVEMENT_NUM = 5
    MAX_BATCH_SIZE = 128
    LIMIT_MEMORY = False
    SEARCH_MAX_ITER = 200



from autokeras.constant import Constant
from autokeras.graph import Graph
from autokeras.layers import StubBatchNormalization, StubConv, StubDropout, StubPooling, StubDense, StubFlatten, \
    StubReLU, StubSoftmax


class ClassifierGenerator:
    def __init__(self, n_classes, input_shape):
        self.n_classes = n_classes
        self.input_shape = input_shape
        if len(self.input_shape) > 4:
            raise ValueError('The input dimension is too high.')
        if len(self.input_shape) < 2:
            raise ValueError('The input dimension is too low.')

    def _get_shape(self, dim_size):
        temp_list = [(dim_size,), (dim_size, dim_size), (dim_size, dim_size, dim_size)]
        return temp_list[len(self.input_shape) - 2]


class DefaultClassifierGenerator(ClassifierGenerator):
    def __init__(self, n_classes, input_shape):
        super().__init__(n_classes, input_shape)

    def generate(self, model_len=Constant.MODEL_LEN, model_width=Constant.MODEL_WIDTH):
        pooling_len = int(model_len / 4)
        graph = Graph(self.input_shape, False)
        temp_input_channel = self.input_shape[-1]
        output_node_id = 0
        for i in range(model_len):
            output_node_id = graph.add_layer(StubReLU(), output_node_id)
            output_node_id = graph.add_layer(StubConv(temp_input_channel, model_width, kernel_size=3), output_node_id)
            output_node_id = graph.add_layer(StubBatchNormalization(model_width), output_node_id)
            output_node_id = graph.add_layer(StubDropout(Constant.CONV_DROPOUT_RATE), output_node_id)
            temp_input_channel = model_width
            if pooling_len == 0 or ((i + 1) % pooling_len == 0 and i != model_len - 1):
                output_node_id = graph.add_layer(StubPooling(), output_node_id)

        output_node_id = graph.add_layer(StubFlatten(), output_node_id)
        output_node_id = graph.add_layer(StubDense(graph.node_list[output_node_id].shape[0], self.n_classes),
                                         output_node_id)
        graph.add_layer(StubSoftmax(), output_node_id)
        return graph
